Promise		= require 'bluebird'
path		= require 'path'
os			= require 'os'
fs			= require 'fs'
parseUrl	= require('url').parse
querystring	= require 'querystring'
Cookies		= require 'cookies'
zlib		= require 'zlib'
colors		= require 'colors'
Busboy		= require 'busboy'
crypto		= require 'crypto'

{ clone, merge, typeOf, toArray } = require './helpers/utils'
{ parseProperties, mimetypes, httpcodes } = require './helpers'

module.exports = class RequestHandler
	constructor: (@req, @res, @lance) ->
		@time			= new Date()
		@code			= 200
		@headers		= { 'content-type': 'text/html; charset=utf-8' }
		@body			= ''
		@encoding		= 'utf8'
		@json			= {}
		@template		= { view: '', data: {}, templater: @lance.templater }
		@redirect		= ''
		@redirectQuery	= {}
		@query			= {}

		@lance.emit 'request.unparsed', this

		@parseRequest().finally =>
			@lance.emit 'request', this
			
			if @route.callback
				@route.callback.apply @lance, [ this ]
			else
				@lance.requestCallback.apply @lance, [ this ]

	###
		Standardizes the serving of templates, json etc. based on the set options in this instance
	###
	serve: ->
		promise = do Promise.coroutine =>
			@lance.emit 'serve', this

			if @redirect
				@serveRedirect @redirect, @redirectQuery or {}
			else
				if @template.view
					@serveTemplate()
				else
					@serveJson()

			yield return

		promise.catch (err) =>
			@lance.emit 'error', err
			@serveHttpCode 500

	serveTemplate: (view = @template.view, data = @template.data) ->
		promise = do Promise.coroutine =>
			@lance.emit 'serveTemplate', this

			if view
				data = merge {
					@query
					o: {
						@time, @code, @headers, @body, @encoding, @json, @template, @redirect
						@redirectQuery, @query, @method, @route, @routes, @path, @splats, @cookies
						@files, @session
					}
				}, data

				rendered = yield @template.templater.render view, data

				@body = rendered

			yield return @respond()

		promise.catch (err) =>
			@lance.emit 'error', err
			@serveHttpCode 500

	serveJson: (json) ->
		@lance.emit 'serveJson', this

		@res.statusCode = @code
		@res.setHeader 'content-type', 'application/json'
		@res.end JSON.stringify json or @json

	serveRedirect: (path = @redirect, query = @redirectQuery) ->
		@lance.emit 'serveRedirect', this

		if hash = path.match( /(#[\w\d_-]+)/i )?[1]
			path = path.replace hash, ''

		if Object.keys( query ).length
			path += '?' + querystring.stringify query

		if hash
			path += hash

		@res.statusCode = 302
		@res.setHeader 'location', path
		@res.end()

	serveHttpCode: (code = @code, body = '', headers = { 'content-type': 'text/plain; charset=utf-8' }) ->
		@lance.emit 'serveHttpCode', this

		@res.statusCode = code
		@res.setHeader key, val for key, val of headers

		if body
			@res.end body
		else
			title = httpcodes[ code.toString() ] or ''
			@res.end "#{code} #{title}"

	respond: ->
		@lance.emit 'respond', this

		@res.statusCode = @code
		@res.setHeader key, val for key, val of @headers

		length = Buffer.byteLength @body.toString(), 'utf8'

		@res.setHeader 'content-length', length or 0
		@res.end @body, @encoding

	compress: (body, done = ->) -> #todo: make it work
		acceptEncoding = @req.headers['accept-encoding'] or ''

		if acceptEncoding.match /\bgzip\b/i
			res.setHeader 'content-encoding', 'gzip'
			zlib.gzip body, handler
		else if acceptEncoding.match /\bdeflate\b/i
			res.setHeader 'content-encoding', 'deflate'
			zlib.deflate body, handler
		else
			res.setHeader 'content-length', body.length or 0
			done null, body

	next: ->
		url = @route.url

		newRoute		= @lance.router.match url.pathname, @method, @route.index + 1
		newRoute.url	= url

		@route		= newRoute
		@path		= newRoute.path
		@splats		= newRoute.splats

		if newRoute.callback
			newRoute.callback this

	parseRequest: ->
		new Promise (resolve) =>
			@method		= @req.method.toUpperCase()

			url			= parseUrl @req.url, true
			@route		= @lance.router.match url.pathname, @method
			@route.url	= url

			@routes		= @lance.router.namedRoutes
			@path		= @route.path
			@splats		= @route.splats

			@cookies	= new Cookies @req, @res

			@files = {}

			switch @method
				when 'GET'
					@query = url.query or {}

					resolve()

				when 'POST'
					@query = {}

					# todo: consider adding events for file upload stuff so we can check progress n stuff
					opt = clone @lance.cfg.server.busboy
					merge opt, { headers: @req.headers }

					try
						@stream = new Busboy opt
					catch err
						return resolve()

					@stream.on 'file', (field, stream, filename, encoding, mimetype) =>
						mimeExt = ''
						for ext, type of mimetypes when type is mimetype
							mimeExt = ext
							break

						if not mimeExt
							@lance.emit 'error', new Error "`#{mimetype}` is an unknown mimetype in file upload", 'Lance.RequestHandler.parseRequest'
							return null

						randomId	= crypto.randomBytes(2).toString('hex')
						newFileName	= new Date().getTime() +  randomId + '.' + mimeExt

						tempFilePath = path.join os.tmpDir(), newFileName
						length = 0

						stream.on 'end', =>
							deleteFile = -> fs.exists tempFilePath, (exists) -> if exists then fs.unlink tempFilePath

							if stream.truncated
								fs.unlink tempFilePath
							else
								timeout = setTimeout deleteFile, 60000

							@query[ field ] =
							@files[ field ] = {
								field
								filename
								stream
								encoding
								mimetype
								ext: mimeExt
								truncated: stream?.truncated
								tempFilePath
								newFileName
								delete: ->
									clearTimeout timeout
									deleteFile()
							}

						stream.pipe fs.createWriteStream tempFilePath

					@stream.on 'field', (field, val, truncatedField, truncatedVal) =>
						if @query[ field ] isnt undefined
							if not typeOf.Array @query[field]
								@query[ field ] = [ @query[ field ] ]
							
							@query[ field ].push val
						else
							@query[ field ] = val

					@stream.on 'finish', resolve

					@req.pipe @stream

				else
					postBody	= ''

					@req.on 'data', (chunk) -> postBody += chunk
					@req.on 'end', =>
						@query = querystring.parse( postBody ) or {}
						resolve()

		.then =>
			@query = parseProperties @query


