path		= require 'path'
Promise		= require 'bluebird'
fs			= Promise.promisifyAll require 'fs'
cluster		= require 'cluster'
uglifyJs	= require 'uglify-js'
mkdirp		= Promise.promisify require 'mkdirp'
Browserify	= require 'browserify'
Coffeeify	= require 'coffeeify'

{ clone, merge, typeOf, isAbsolute, changeExt, isExt, changeExt, exploreDir, empty, minify, minifyCss, escapeRegExp, setInterval, setTimeout } = require './helpers/utils'
{ promisify } = require './helpers'

module.exports = Templater = ->
	self = this

	@constructor = (newCfg, @lance) =>
		oldCfg = require('../cfg').templater

		@cfg = clone.json oldCfg

		merge @cfg, newCfg if newCfg

		@cfg.root = @cfg.root or @lance.paths.root

		@locals = {}

		{ @stylus, @coffee, @templater, @css, @js, @assets, ect } = @cfg
		
		@coffee.engine		= @coffee.engine or require 'coffee-script'
		try @stylus.engine	= @stylus.engine or require 'stylus'

		try
			if not @templater and ect
				@templater		= ect
				@templater.name	= 'ect'

				if not @templater.engine
					try @templater.engine = require 'ect'

			if @templater.name is 'ect'
				@constructEct @templater
		catch err
			throw new Error 'Invalid templating engine'

		if @css.inherit
			@css.findIn or @css.findIn = @stylus.findIn
			@css.saveTo or @css.saveTo = @stylus.saveTo

		if @js.inherit
			@js.findIn or @js.findIn = @coffee.findIn
			@js.saveTo or @js.saveTo = @coffee.saveTo

		if @stylus.inherit
			@stylus.findIn or @stylus.findIn = @css.findIn
			@stylus.saveTo or @stylus.saveTo = @css.saveTo

		if @coffee.inherit
			@coffee.findIn or @coffee.findIn = @js.findIn
			@coffee.saveTo or @coffee.saveTo = @js.saveTo

		@css.match		= @css.ext if not @css.match
		@js.match		= @js.ext if not @js.match
		@stylus.match	= @stylus.ext if not @stylus.match
		@coffee.match	= @coffee.ext if not @coffee.match

		if @cfg.merge then for item in @cfg.merge
			if not item.files and item.file
				item.files = item.file
				delete item.file

			if not typeOf.Array item.files
				item.files = [ item.files ]

		@cache =
			toCss:
				watching	: {}
				renderedAt	: {}
			toJs:
				watching	: {}
				renderedAt	: {}
			merge:
				toJs	: {}
				toCss	: {}
				watching: {}

			bundle: {
				watching: {}
			}

			watchTimeouts: {}
			writeTimeouts: {}

		@ignoreUnmerged = @cfg.merge and @cfg.mergeOnly

		# Backwards compat
		if @cfg.merge
			for group in @cfg.merge
				group.saveTo = group.name if not group.saveTo and group.name

		if @cfg.autoInitialize
			@initializer = @initialize().then()
		else
			@initializer = new Promise (resolve) -> resolve()
	
		return this

	# finds, renders and watches files according to the config
	@initialize = ==>
		# Causes css, js, coffee and stylus to inherit findIn and saveTo
		# paths from one another should they not be set

		await = []
		
		if @js.render and @js.findIn
			await.push if @js.watch
				@find.js.renderAndWatch()
			else
				@find.js.render()

		if @coffee.render and @coffee.engine
			await.push if @coffee.watch
				@find.coffee.renderAndWatch()
			else
				@find.coffee.render()
		
		if @css.render and @css.findIn
			await.push if @css.watch
				@find.css.renderAndWatch()
			else
				@find.css.render()

		if @stylus.render and @stylus.engine
			await.push if @stylus.watch
				@find.stylus.renderAndWatch()
			else
				@find.stylus.render()

		if @assets.render and @assets.findIn
			if @assets.watch
				@find.assets.renderAndWatch().then()
			else
				@find.assets.render().then()

		return Promise.all await

	@constructEct = (templater) =>
		if typeOf.Function templater.engine
			templater.options.root = if templater.findIn and typeOf.String templater.findIn
				@resolveDirectory templater.findIn
			else
				templater.options.root or @cfg.root

			templater.options.ext	= templater.options.ext or templater.ext
			templater.engine		= templater.engine templater.options

		templater.ext = templater.ext or templater.options.ext or '.ect'

	#
	# Render
	#


	###
		Alias to `@render.template`.
	###
	@render = (dir, locals) =>
		return @render.template dir, locals

	###
		Renders a template into memory.
	###
	@render.template = (dir = '', locals) ==>
		@lance.emit 'templater.render.template', dir, locals

		if locals
			locals = merge.black locals, clone.merge( @locals ), 6

		rendered = yield new Promise (resolve, reject) =>
			@templater.engine.render dir, locals, (err, rendered = '') =>
				return reject err if err
				resolve rendered

		if @templater.minify and rendered
			rendered = minify rendered

		return rendered

	@render.toCss = (fileDir) ==>
		return if isExt fileDir, @stylus.ext
			@render.stylus fileDir
		else if isExt fileDir, @css.ext
			@render.css fileDir
		else
			null

	@render.toJs = (fileDir) ==>
		return if isExt fileDir, @coffee.ext
			@render.coffee fileDir
		else if isExt fileDir, @js.ext
			@render.js fileDir
		else
			null

	@render.css = (fileDir) ==>
		fileDir = @resolveDirectory fileDir

		@lance.emit 'templater.render.css', fileDir

		if not ( yield @exists fileDir )
			throw new Error "#{fileDir} doesnt exist"

		file = yield @readFile fileDir

		newFileDir = @createSaveToPath fileDir, @css

		if @css.minify
			file = minifyCss file

		return if @render.toCss.merge.add fileDir, file
			@render.toCss.merge()
		else if not @ignoreUnmerged
			@writeFile newFileDir, file
		else
			null

	@render.stylus = (fileDir) ==>
		fileDir = @resolveDirectory fileDir

		@lance.emit 'templater.render.stylus', fileDir

		if not ( yield @exists fileDir )
			throw new Error "#{fileDir} doesnt exist"

		file = yield @readFile fileDir
	
		promise = new Promise (resolve, reject) =>
			try
				engine = @stylus.engine file, @stylus.options
				engine.set 'paths', [ @cfg.root, path.dirname fileDir ]
				dependencies = engine.deps file
			catch err
				err.reverse = true
				@lance.emit 'error', err
				return reject err

			engine.render (err, rendered = '') =>
				return reject err if err
				return resolve rendered if not rendered

				newFileDir = @createSaveToPath fileDir, @stylus

				if @stylus.minify
					rendered = minifyCss rendered

				if dependencies
					# For all dependencies, watch the files and re-render the root file on change
					for dep in dependencies then do (dep) =>
						@watch.one dep, => @render.stylus fileDir

				if @render.toCss.merge.add fileDir, rendered
					resolve @render.toCss.merge()
				else if not @ignoreUnmerged
					resolve @writeFile newFileDir, rendered
				else
					resolve()

		promise.catch (err) -> console.error err?.stack or err


	@render.js = (fileDir) ==>
		fileDir = @resolveDirectory fileDir

		@lance.emit 'templater.render.js', fileDir

		if not ( yield @exists fileDir )
			throw new Error "#{fileDir} doesnt exist"

		rendered = yield @readFile fileDir

		newFileDir = @createSaveToPath fileDir, @js

		if @js.minify
			try
				if result = uglifyJs.minify rendered, { fromString: true, mangle: false }
					rendered = result.code

		return if @render.toJs.merge.add fileDir, rendered
			@render.toJs.merge()
		else if not @ignoreUnmerged
			@writeFile newFileDir, rendered
		else
			null

	@render.coffee = (fileDir) ==>
		fileDir = @resolveDirectory fileDir

		@lance.emit 'templater.render.coffee', fileDir

		if not ( yield @exists fileDir )
			throw new Error "#{fileDir} doesnt exist"

		file = yield @readFile fileDir

		rendered = @coffee.engine.compile file

		newFileDir = @createSaveToPath fileDir, @coffee

		if @coffee.minify
			try
				if result = uglifyJs.minify rendered, { fromString: true, mangle: false }
					rendered = result.code

		return if @render.toJs.merge.add fileDir, rendered
			@render.toJs.merge()
		else if not @ignoreUnmerged
			@writeFile newFileDir, rendered
		else
			null

	@render.asset	=
	@render.assets	= (fileDir) ==>
		fileDir = @resolveDirectory fileDir

		@lance.emit 'templater.render.assets', fileDir

		if not ( yield @exists fileDir )
			throw new Error "#{fileDir} doesnt exist"

		newFileDir = @createSaveToPath fileDir, @assets
		readStream = fs.createReadStream fileDir

		return @writeFileStream readStream, newFileDir

	@render.toCss.merge = (fileDir) =>
		@render.merge @cache.merge.toCss, @css

	@render.toCss.merge.add	= (fileDir, str) =>
		model = if isExt fileDir, @stylus.ext then @stylus else @css

		@render.merge.add @cache.merge.toCss, model, fileDir, str

	@render.toCss.renderedMerge = (fileDir) ==>
		@render.renderedMerge @cache.merge.toCss, @css, @render.toCss

	@render.toJs.merge = (fileDir) =>
		@render.merge @cache.merge.toJs, @js

	@render.toJs.merge.add	= (fileDir, str) =>
		model = if isExt fileDir, @coffee.ext then @coffee else @js
		@render.merge.add @cache.merge.toJs, model, fileDir, str

	@render.toJs.renderedMerge = (fileDir) ==>
		@render.renderedMerge @cache.merge.toJs, @js, @render.toJs

	@render.renderedMerge = (mergeCache, model, renderer) ==>
		await = []
		for key, group of mergeCache
			for fileDir of group
				await.push renderer fileDir

		yield Promise.all await

		return @render.merge mergeCache, model


	@render.bundle = (destination, source) ==>
		source = [ source ] if not typeOf.Array source

		for file, index in source
			model = @resolveModel file
			source[index] = @resolveDirectory file, @resolveDirectory model.findIn

		console.log source
		
		#files[index] = fs.createReadStream file for file, index in files
		
		b = Browserify source, @cfg.browserify
		b.on 'file', (file) =>
			@watch file, =>
				

		b.transform Coffeeify
		
		readStream = b.bundle()

		saveTo = @createSaveToPath destination

		return yield @writeFileStream readStream, saveTo

	# render multiple bundles
	@render.bundles = (bundles = @cfg.bundles) ->

	# not called on its own. instead, @render.[coffee|stylus].merge
	# supplies it with mergeCache and a file extension
	@render.merge = (mergeCache, model) ==>
		return null if not @cfg.merge

		@lance.emit 'templater.render.merge', model

		{ ext, saveTo, findIn } = model

		findIn = [ findIn ] if not typeOf.Array findIn

		return null if not saveTo or not findIn or empty mergeCache

		await = []
		for group in @cfg.merge
			continue if group.saveTo not of mergeCache

			merged = ''
			mergedFiles = []

			for relPath in group.files
				if not typeOf.RegExp relPath
					relPath = new RegExp  '^\/?' + ( escapeRegExp ( relPath.replace /^\.?\//, '' ) ) + '$'

				for fileDir, rendered of mergeCache[ group.saveTo ] when fileDir not in mergedFiles
					for findInPath in findIn
						constructedDir = new RegExp '^' + escapeRegExp @resolveDirectory path.join ( findInPath or '' )
						dir = fileDir.replace constructedDir, ''

						if relPath.test dir
							merged += "/* #{ path.basename fileDir } */\n#{ rendered }\n\n"
							mergedFiles.push fileDir
							break

			saveTo		= @resolveDirectory saveTo
			newFileDir	= @createSaveToPath group.saveTo, model

			@lance.emit 'templater.render.merge merged', mergedFiles, group

			await.push @deferredWrite 'merge:' + group.saveTo, newFileDir, merged

		return Promise.all await

	@render.merge.add = (cache, model, fileDir, file) =>
		return false if not fileDir or not file? or not @cfg.merge

		@lance.emit 'templater.render.merge.add', fileDir, model

		{ findIn } = model
		findIn = [ findIn ] if not typeOf.Array findIn

		added = false

		for group in @cfg.merge
			found = false

			@lance.emit 'templater.render.merge.add search', fileDir, group

			for relPath in group.files
				if not typeOf.RegExp relPath
					relPath = new RegExp  '^\/?' + ( escapeRegExp ( relPath.replace /^\.?\//, '' ) ) + '$'

				for findInPath in findIn
					constructedDir = new RegExp '^' + escapeRegExp @resolveDirectory path.join ( findInPath or '' )
					dir = fileDir.replace constructedDir, ''

					@lance.emit 'templater.render.merge.add matching', fileDir, relPath, dir

					if relPath.test dir
						if not cache[ group.saveTo ]
							cache[ group.saveTo ] = {}

						cache[ group.saveTo ][ fileDir ] = file

						@lance.emit 'templater.render.merge.add matched', fileDir, group

						found = added = true
						break

				break if found

		if not added
			@lance.emit 'templater.render.merge.add unmatched', fileDir

		return added

	#
	# Find
	#


	###
		Iterates over a directory and matching files.
	###
	@find = (findIn = '', match = '', next = ->) ==>
		findIn	= [ findIn ] if not typeOf.Array findIn
		
		if typeOf.String match
			matchFn = (fileDir = '') -> return isExt fileDir, match
		else if typeOf.RegExp match
			matchFn = (fileDir = '') -> return match.test fileDir
		else if typeOf.Function match
			matchFn = match
		else
			return null

		await = []
		for dir in findIn
			dir = @resolveDirectory dir

			@lance.emit 'templater.find', dir, match

			await.push exploreDir dir, { depth: @cfg.depth or 8 }, (file, fileDir, fileName, name, fileExt = '') =>
				@lance.emit 'templater.find.file', file, fileDir, fileName, name, fileExt

				if matchFn fileDir
					@lance.emit 'templater.find.fileMatch', file, fileDir, fileName, name, fileExt

					return next file, fileDir, fileName, name, fileExt

		return Promise.all await

	# Find - CSS

	@find.css = (findIn, next) =>
		findIn = findIn or @css.findIn
		@find findIn, @css.match, next

	@find.css.watch = (findIn) =>
		@find.css findIn, (file, fileDir) =>
			@watch.toCss fileDir

	@find.css.render = (findIn) ==>
		await = []
		yield @find.css findIn, (file, fileDir) =>
			await.push @render.toCss fileDir

		return Promise.all await

	@find.css.renderAndWatch = (findIn) ==>
		await = []
		yield @find.css findIn, (file, fileDir) =>
			await.push @render.toCss fileDir
			@watch.toCss fileDir

		return Promise.all await

	# Find - Stylus

	@find.stylus = (findIn, next) =>
		findIn = findIn or @stylus.findIn
		@find findIn, @stylus.match, next

	@find.stylus.watch = (findIn) =>
		@find.stylus findIn, (file, fileDir) =>
			@watch.toCss fileDir

	@find.stylus.render = (findIn) ==>
		await = []
		yield @find.stylus findIn, (file, fileDir) =>
			await.push @render.stylus fileDir

		return Promise.all await

	@find.stylus.renderAndWatch = (findIn) ==>
		await = []
		yield @find.stylus findIn, (file, fileDir) =>
			await.push @render.stylus fileDir
			@watch.toCss fileDir

		return Promise.all await

	# Find - Js

	@find.js = (findIn, next) =>
		findIn = findIn or @js.findIn
		@find findIn, @js.match, next

	@find.js.watch = (findIn) =>
		@find.js findIn, (file, fileDir) =>
			@watch.toJs fileDir

	@find.js.render = (findIn = '') ==>
		await = []
		yield @find.js findIn, (file, fileDir) =>
			await.push @render.toJs fileDir

		return Promise.all await

	@find.js.renderAndWatch = (findIn = '') ==>
		await = []
		yield @find.js findIn, (file, fileDir) =>
			await.push @render.toJs fileDir
			@watch.toJs fileDir

		return Promise.all await

	# Find - Coffeescript

	@find.coffee = (findIn, next) =>
		findIn = findIn or @coffee.findIn
		@find findIn, @coffee.match, next

	@find.coffee.watch = (findIn) =>
		@find.coffee findIn, (file, fileDir) =>
			@watch.toJs fileDir

	@find.coffee.render = (findIn) ==>
		await = []
		yield @find.coffee findIn, (file, fileDir) =>
			await.push @render.coffee fileDir

		return Promise.all await

	@find.coffee.renderAndWatch = (findIn) ==>
		await = []
		yield @find.coffee findIn, (file, fileDir) =>
			await.push @render.coffee fileDir
			@watch.toJs fileDir

		return Promise.all await

	# Find - Assets

	@find.assets = (findIn, next) =>
		findIn = findIn or @assets.findIn
		@find findIn, @assets.match, next

	@find.assets.watch = (findIn) =>
		@find.assets findIn, (file, fileDir) =>
			@watch.asset fileDir

	@find.assets.render = (findIn) ==>
		await = []
		yield @find.assets findIn, (file, fileDir) =>
			await.push @render.assets fileDir

		return Promise.all await

	@find.assets.renderAndWatch = (findIn) ==>
		await = []
		yield @find.assets findIn, (file, fileDir) =>
			await.push @render.assets fileDir
			@watch.asset fileDir

		return Promise.all await

	@watch = (fileDir, callback) =>
		return false if not fileDir

		@lance.emit 'templater.watch', fileDir

		fs.watch fileDir, (event) =>
			return false if event isnt 'change'

			watchTimeouts = @cache.watchTimeouts

			if fileDir of watchTimeouts
				clearTimeout watchTimeouts[ fileDir ]
			else
				@lance.emit 'templater.watch.change', fileDir, event
				@lance.emit "templater.watch.change.#{fileDir}", fileDir, event

			watchTimeouts[ fileDir ] = setTimeout @cfg.watchTimeout, =>
				delete watchTimeouts[ fileDir ]
				callback fileDir, event if callback

		return true

	@watch.one = (fileDir, callback) =>
		if fileDir not of @cache.watching
			if @watch fileDir, callback
				@cache.watching[ fileDir ] = true

	@watch.toCss = (fileDir) =>
		renderer = if isExt fileDir, @stylus.ext then @render.stylus else @render.toCss
		@watch.one fileDir, -> renderer fileDir

	@watch.toJs = (fileDir) =>
		renderer = if isExt fileDir, @coffee.ext then @render.coffee else @render.toJs
		@watch.one fileDir, -> renderer fileDir

	@watch.asset	=
	@watch.assets	= (fileDir) =>
		@watch.one fileDir, => @render.asset fileDir

	#
	# Files
	#

	@exists = (fileDir) => new Promise (resolve) -> fs.exists fileDir, resolve

	@deferredWrite = (key, fileDir, file) =>
		return new Promise (resolve) =>
			if cache = @cache.writeTimeouts[ key ]
				cache.resolve()
			else
				cache = @cache.writeTimeouts[ key ] = {}
			
			cache.fulfill = =>
				@writeFile fileDir, file

			cache.resolve = resolve

			@refreshDeferredWrite key

	@refreshDeferredWrite = (key, time = @cfg.writeTimeout) =>
		if cache = @cache.writeTimeouts[ key ]
			clearTimeout cache.timeout if cache.timeout

			cache.timeout = setTimeout time, =>
				cache.resolve @cache.writeTimeouts[ key ].fulfill()
				delete @cache.writeTimeouts[ key ]

	@resolveDirectory = (dir = '', root = '') ->
		return dir if typeof dir isnt 'string'

		return if isAbsolute dir
			dir
		else
			path.join root or @cfg.root or '', dir

	@readFile = (fileDir, encoding) ==>
		if not encoding and encoding isnt false
			encoding = 'utf8'

		return fs.readFileAsync fileDir, encoding

	@writeFile = (fileDir, file) ==>
		@lance.emit 'templater.writeFile', fileDir

		yield @createDirectory fileDir

		return yield fs.writeFileAsync fileDir, file

	@writeFileStream = (readStream, fileDir) ==>
		@lance.emit 'templater.writeFile', fileDir

		yield @createDirectory fileDir

		console.log fileDir

		writeStream = fs.createWriteStream fileDir
		return new Promise (resolve, reject) =>
			errored = false
			error = (err) ->
				reject err if not errored
				errored = true

			readStream.once 'error', error
			writeStream.once 'error', error

			readStream.pipe writeStream
				.on 'error', error
				.on 'close', resolve

	@createDirectory = (fileDir) ==>
		return mkdirp path.dirname( fileDir ), @cfg.writePermissions or ( 0o0755 & ( ~process.umask() ) )

	# By default, this preserve findIn rooted paths into the saveTo directory. Currently unchangable.
	@createSaveToPath = (fileDir, model) ->
		if not model
			model = @resolveModel fileDir

		{ findIn, saveTo, ext } = model

		findIn = findIn[0] if typeOf.Array findIn

		findIn	= @resolveDirectory findIn
		saveTo	= @resolveDirectory saveTo
		fileDir	= @resolveDirectory fileDir, findIn

		# todo: match the findIn array with the fileDir first?
		findIn = new RegExp ( '^' + escapeRegExp findIn )

		partial = if @cfg.preserveDirectory
			fileDir.replace findIn, ''
		else
			path.basename fileDir

		newFileDir = path.join saveTo, partial

		if ext
			newFileDir = changeExt newFileDir, ext

		return newFileDir

	@isToJs = (fileDir) => return isExt( fileDir, @coffee.ext ) or isExt( fileDir, @js.ext )
	@isToCss = (fileDir) => return isExt( fileDir, @stylus.ext ) or isExt( fileDir, @css.ext )
	@resolveModel = (fileDir) =>
		return switch
			when isExt fileDir, @coffee.ext then @coffee
			when isExt fileDir, @js.ext then @js
			when isExt fileDir, @stylus.ext then @stylus
			when isExt fileDir, @css.ext then @css
			when isExt fileDir, @templater.ext then @templater

	promisify.all.generators this

	return @constructor.apply this, arguments

