Promise = require 'bluebird'

module.exports = helpers = {}

{ merge, clone, typeOf } = helpers.utils = require './helpers/utils'

#
# Workflow
#


###
	Iterates over all properties of an object and if
	a function is found, apply Promise.coroutine or Promise.method
	depending on whether it's a generator or not.

	@param obj {Object}
	@param whiteList {Array} Keys in `obj`
	@return obj
###
helpers.promisify = (obj, whiteList = []) ->
	for key, val of obj
		continue if whiteList.length and key not in whiteList

		if typeOf.Function val
			obj[key] = helpers.promisify.create val

	return obj

helpers.promisify.create = (val) ->
	if val.constructor.name is 'GeneratorFunction'
		return Promise.coroutine val
	else
		return Promise.method val

helpers.promisify.createCoroutine = (val) ->
	if val.constructor.name is 'GeneratorFunction'
		return Promise.coroutine val

	return val

helpers.promisify.createMethod = (val) ->
	if val.constructor.name isnt 'GeneratorFunction'
		return Promise.method val

	return val

# Only promisifies generator functions in an object
helpers.promisify.generators = (obj, whiteList = []) ->
	for key, val of obj
		continue if whiteList.length and key not in whiteList

		if typeOf.Function val
			obj[key] = helpers.promisify.createCoroutine val

	return obj

# Only promisifies NON-generator functions in an object
helpers.promisify.functions = (obj, whiteList = []) ->
	for key, val of obj
		continue if whiteList.length and key not in whiteList

		if typeOf.Function val
			obj[key] = helpers.promisify.createMethod val

	return obj

helpers.promisify.all = (obj, depth = 20, promisifier = helpers.promisify.create) ->
	if --depth > 1
		for key, val of obj
			switch typeOf val
				when 'function'
					val		= helpers.promisify.all val, depth, promisifier
					newFn	= promisifier val

					merge newFn, val, null, [ 'object', 'function' ]
					obj[key] = newFn
				when 'object'
					helpers.promisify.all val, depth, promisifier
	
	return obj

helpers.promisify.all.generators = (obj, depth) ->
	return helpers.promisify.all obj, depth, helpers.promisify.createCoroutine

helpers.promisify.all.functions = (obj, depth) ->
	return helpers.promisify.all obj, depth, helpers.promisify.createMethod

#
# Formatting
#


helpers.parseProperties	=
helpers.transpose		= require './helpers/transposer'

###
	All arrays and objects within the object/array become json.

	@param obj {Object, or Array}
	@return obj
###
helpers.jsonifyProperties = (obj) ->
	for key, val of obj
		switch typeOf val
			when 'object', 'array'
				obj[key] = JSON.stringify val

	return obj


helpers.prettyErrorStack	= require './helpers/prettyErrorStack'
helpers.httpcodes			= require './helpers/httpcodes'
helpers.mimetypes			= require './helpers/mimetypes'
helpers.events				= require './helpers/events'
helpers.sourceMaps			= require './helpers/sourceMaps'







